//
//  PsuedokoSolve.m
//  Pseudoko
//
//  Created by David Phillip Oster on 1/9/07.
//  Copyright 2007 David Phillip Oster. 
//

#import "PsuedokoSolve.h"
#import "PuzzleProxy.h"
#import "PseudokoMove.h"

typedef struct MyCell {
    int column;
    int row;
} MyCell;


@interface PseudokoDocument (SolvePrivate)
- (void)showPuzzleRegisterForUndo;
- (void)nextMove;
- (void)finishSolve;
- (BOOL)isBlockedIntValue:(int)val row:(int)row column:(int)column;
- (BOOL)isBlockedIntValue:(int)val row:(int)row;
- (BOOL)isBlockedIntValue:(int)val column:(int)column;
- (BOOL)isBlockedSubGridIntValue:(int)val row:(int)row column:(int)column;
- (BOOL)isSolved;
- (void)setMoves:(NSMutableArray *)aMoves;
@end

@implementation PseudokoDocument (Solve)

- (void)solve:(id)sender {
  if ([self endEditing]) {
    [[self undoManager] beginUndoGrouping];
    [self setIsProvenNotSolvable:NO];
    [self showPuzzleRegisterForUndo];  // and record for undo.
    [self nextMove];
  }
}

@end

@implementation PseudokoDocument (SolvePrivate)
// copy it to itself using the KVC compatible setter, so bindings will update the U.I.
- (void)showPuzzleRegisterForUndo{
  for(int  i = 0; i < 9*9; ++i) {
    PuzzleProxy *proxy = [puzzleProxys objectAtIndex:i];
    [proxy setValue: [proxy value]];
  }
  [[[self undoManager] prepareWithInvocationTarget:self] setIsProvenNotSolvable:[self isProvenNotSolvable]];
  [[[self undoManager] prepareWithInvocationTarget:self] setMoves:[[moves mutableCopy] autorelease]];
}

- (MyCell)findEmptyCellAfterRow:(int)row column:(int)column{
  MyCell cell = {-1, -1};
  for(int  j = row; j < 9; ++j) {
    for(int  i = column; i < 9; ++i) {
      if (0 == puzzle[j][i]) {
        cell.row = j;
        cell.column = i;
        return cell;
      }
    }
    column = 0;
  }
  return cell;
}

// given the current state of the board, return a candidate next move.
// nil if no moves.
- (PseudokoMove *)generateNextMove {
  PseudokoMove *move = nil; // will be return value.
  MyCell cell;
  cell.row = -1;
  cell.column = 9;  // incrementing 'cell' puts us at the start of the board
  PseudokoMove *lastMove = nil; // used by backtracker.
  for(;;) {
    int newValue;
    if (lastMove) {
      newValue = 1+[lastMove intValue];
      cell.row = [lastMove row];
      cell.column = [lastMove column];
      [lastMove release];
      lastMove = nil;
    } else if (0 != [moves count]) {
      PseudokoMove *lastMove = [moves objectAtIndex:[moves count] - 1];
      cell = [self findEmptyCellAfterRow:lastMove->row column:lastMove->column];
      if (cell.row != -1) {
        newValue = 1+((puzzle[cell.row][cell.column])%8);
      }else{
        return nil;
      }
    }else{
      cell = [self findEmptyCellAfterRow:cell.row column:cell.column];
      if (cell.row != -1) {
        newValue = 1+((puzzle[cell.row][cell.column])%8);
      }else{
        return nil;
      }
    }
    for (;newValue <= 9;newValue++){
      if ( ! [self isBlockedIntValue:newValue row:cell.row column:cell.column]) {
        move = [[[PseudokoMove alloc] initWithValue:newValue 
          previousValue:puzzle[cell.row][cell.column] 
                    row:cell.row
                 column:cell.column] autorelease];
        return move;  // <-- normal loop exit
      }
    }

    // no moves with the board in its current state. backtrack until we can move.
    if (0 <  [moves count]) {
      while (0 < [moves count]) {
        [lastMove release];
        lastMove = [[moves objectAtIndex:[moves count] - 1] retain];
        [lastMove unapply:self];
        [moves removeLastObject];
        if ([lastMove intValue] < 9) {
          break;
        }
      }
    }else{
      [self setIsProvenNotSolvable:YES];
      return nil; // <-- giving up
    }
  }
  return move;
}

// generate and apply moves until done.
- (void)nextMove {
  PseudokoMove *move = [self generateNextMove];
  while (move) {
    [moves addObject:move];
    [move apply:self];
    move = [self generateNextMove];  // TODO: this could fire on a timer to animate the solution process.
  }
  if ( ! move) {
    [self finishSolve];
  }
}

// cleanup at end of solve - finalize the undo.
- (void)finishSolve {
  [self showPuzzleRegisterForUndo];  // and reveal the state change.
  [[self undoManager] setActionName:NSLocalizedString(@"Solve", @"")];
  if ([self isSolved]) {
    [self setIsProvenNotSolvable:NO];
  }
  [[self undoManager] endUndoGrouping];
}

// Assumes val non-zero. Helper for isBlockedIntValue:row:column:
- (BOOL)isBlockedIntValue:(int)val row:(int)row{
  for(int i = 0;i < 9; ++i){
    if(val == (puzzle[row][i] & kValueMask)) {
      return YES;
    }
  }
  return NO;
}

// Assumes val non-zero. Helper for isBlockedIntValue:row:column:
- (BOOL)isBlockedIntValue:(int)val column:(int)column{
  for(int j = 0;j < 9; ++j){
    if(val == (puzzle[j][column] & kValueMask)) {
      return YES;
    }
  }
  return NO;
}

// Assumes val non-zero. Helper for isBlockedIntValue:row:column:
- (BOOL)isBlockedSubGridIntValue:(int)val row:(int)row column:(int)column{
  int jBase = (row/3) * 3;    // i.e.: 0, 3, 6
  int iBase = (column/3) * 3;
  for(int j = jBase;j < jBase+3; ++j){
    for(int i = iBase;i < iBase+3; ++i){
      if(val == (puzzle[j][i] & kValueMask)) {
        return YES;
      }
    }
  }
  return NO;
}

// return NO if we can't put this value in this spot.
- (BOOL)isBlockedIntValue:(int)val row:(int)row column:(int)column {
  if (0 == val) {
    return NO;
  }
  return [self isBlockedIntValue:val row:row] ||
    [self isBlockedIntValue:val column:column] ||
    [self isBlockedSubGridIntValue:val row:row column:column];
}

- (BOOL)isSolved {
  for(int  j = 0; j < 9; ++j) {
    for(int  i = 0; i < 9; ++i) {
      if (0 == puzzle[j][i]) {
        return NO;
      }
    }
  }
  return YES;
}

- (void)setMoves:(NSMutableArray *)aMoves {
  [moves autorelease];
  moves = [aMoves retain];
}

@end
/*
Permission is hereby granted, free of charge, to any person obtaining
a copy of these files and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
